// VhdlCohen Publishing, copyright 2006 
// wwww.abv-sva.org   ben@abv-sva.org  01/05/06  
// NAME: fifo_rtl
// PURPOSE:  FIFO RTL model 
// 
module fifo (input clk, input reset_n, fifo_if.fslave_if_mp f_if,
	     fifo_csr_if.fifo_ctl_mp f_csr_if
	     ); 
  timeunit 1ns;
  timeprecision 100ps;
  import fifo_pkg::*;
  logic [4 : 0] wr_addr; // write fifo address // bc // changed size from 3 to 4
  logic [4 : 0] rd_addr; // read fifo address  
  typedef enum {NONE, PUSH, POP, PSPP} push_pop_t;
  buffer_t buffer; // fifo storage
  parameter shiftVal = int'(2**BIT_DEPTH);

  logic [4:0] fifo_level, csr_almost_empty, csr_almost_full;
  
  // Push on full with no pop error 
  property p_push_error; 
    @ (posedge clk) 
       not (f_if.push && f_if.full && !f_if.pop); 
  endproperty : p_push_error
  ap_push_error_1 : assert property (p_push_error);
  // mp_push_error : assume property (p_push_error);
  
  // No pop on empty 
  property p_pop_error; 
   @ (posedge clk) 
       not (f_if.pop && f_if.empty); 
  endproperty : p_pop_error
  ap_pop_error_1 : assert property (p_pop_error);
 // mp_pop_error : assume property (p_pop_error);
  
  // RTL Detailed design
  push_pop_t           push_pop;
  // Defining control for maintenance of FIFO word count
  always @*
    begin
      if      (!f_if.push && !f_if.pop) push_pop = NONE;
      else if ( f_if.push && !f_if.pop) push_pop = PUSH;
      else if (!f_if.push &&  f_if.pop) push_pop =POP;
      else if ( f_if.push &&  f_if.pop) push_pop = PSPP;
    end

  // providing READ data to output
  always @ (posedge clk)
     f_if.data_out <= buffer[rd_addr];

  // purpose: writes data from FIFO and controls
  //          words within the FIFO.  It also provides the READ and
  //          WRITE Fifo Addresses. 
 
  always @ (posedge clk or negedge reset_n) 
    begin 
    if (!reset_n)  
        begin  // asynchronous reset (active low)
          rd_addr         <= 0;
          wr_addr         <= 0;
	  fifo_level <= 'b0;
	end
      else
        begin
          case (push_pop)
             NONE :
                     ;              // no push, no pop
            
             POP  : begin                   // no push, pop for READ
	        	if (!f_if.empty)  begin : if_not_emp   // bc -- increment only if OK 
                      rd_addr         <= (rd_addr+1); //  % shiftVal; // bc
	        	      fifo_level <= fifo_level - 1;
	        	end : if_not_emp
              end   
             PUSH :  begin                  // push for WRITE, no pop
               if (!f_if.full) begin : if_not_full // bc -- increment only if OK 
                 wr_addr         <= (wr_addr + 1); // % shiftVal; // bc 
                 buffer[wr_addr] <= f_if.data_in;
	        	 fifo_level <= fifo_level + 1;
	           end : if_not_full
              end
              PSPP :                    // push for WRITE, pop for READ
			   begin
   				 if (!f_if.empty)      // bc -- increment only if OK 
                  rd_addr         <= (rd_addr+1); //  % shiftVal; // bc
                 // if (!f_if.full)       // This is an error detected by assertion 
                 wr_addr         <= (wr_addr + 1); // % shiftVal; // bc 
                 buffer[wr_addr % shiftVal] <= f_if.data_in;  // bc 
              end
            default  : begin
              a_illegal_fifo_cmd_1 : assert (1'b0) else
   		          $warning ("%0t %0m Meta value detected in FIFO command, push_pop %2b ",
                            $time, push_pop);
           end
           
           endcase
        end //else
    end //always

  // Reporting of flags
  // properties used as guide for this design since they define
  // the requirements
  
  always @ (posedge clk)
  begin
    f_if.error <= (f_if.pop && f_if.empty) || 
		  (f_if.push && f_if.full && !f_if.pop);
  end // always
  
  assign f_if.full = (wr_addr[3:0] == rd_addr[3:0]) &&
		     (wr_addr[4] != rd_addr[4]);  // bc update
  assign f_if.empty = (wr_addr == rd_addr);

  always_ff @ (posedge clk) begin : csr_rd_wr
    if (!reset_n) begin
      csr_almost_empty <= 4;  // setting the initial levels 
      csr_almost_full <= 4;   // setting the initial levels 
    end 
    else begin  : else1
      if (f_csr_if.csr_wr) begin : elseif1
	    if (f_csr_if.csr_addr == ALM_EMPTY_REG) begin
	      csr_almost_empty <= f_csr_if.csr_data; // set the configuration 
  	    end 
      else if (f_csr_if.csr_addr == ALM_FULL_REG) begin : elseif2
	    csr_almost_full <= f_csr_if.csr_data;
	  end : elseif2
      else begin : else3
	    a_il_csr_wr_access : assert (1'b0) 
                              else $display ("Illegal CSR WR address %0x ", 
                                             f_csr_if.csr_addr);
	  end : else3
      end : elseif1

      if (f_csr_if.csr_rd) begin : if_rd1
  	    if (f_csr_if.csr_addr == ALM_EMPTY_REG) begin 
	      f_csr_if.csr_data <= csr_almost_empty;
	    end else if (f_csr_if.csr_addr == ALM_FULL_REG) begin
	     f_csr_if.csr_data <= csr_almost_full;
	    end else begin
	      a_il_csr_rd_access : assert (1'b0) 
                               else $display ("Illegal CSR RD address %0x ", 
                                               f_csr_if.csr_addr);
	    end
       end  : if_rd1
    end : else1
  end : csr_rd_wr
  
  assign f_if.almost_empty = (fifo_level <= csr_almost_empty);
  assign f_if.almost_full = (fifo_level >= csr_almost_full);
  
endmodule : fifo



